from abc import ABC, abstractmethod
from typing import Any
import random

class Request(ABC):

    def __init__(self, payload: str) -> None:
        chars = "01234567890abcdef"
        self._id = ''.join(random.choices(chars, k=8))
        self._payload = payload

    @abstractmethod
    def parse_response(self, response: str):
        pass

    def id(self) -> str:
        return self._id

    def payload(self) -> str:
        return self._payload

    def create_request(self, ip: str) -> str:
        return "req:" + ip + ':' + self._id + ':' + self._payload

class CheckAlive(Request):

    def __init__(self) -> None:
        super().__init__("checkAlive()")

    def parse_response(self, response: str) -> str:
        spl = response.split(":")
        clu_serial_hex = spl.pop()
        return int(clu_serial_hex, base=16)

class SetRequest(Request):

    def __init__(self, device_id: str, index: int, val: Any) -> None:
        if isinstance(val, str):
            val = '"' + val + '"'
        elif isinstance(val, bool):
            val = str(val).lower()

        payload = "{}:set({},{})".format(device_id, index, val)
        super().__init__(payload)

    def parse_response(self, response: str):
        return None

class GetRequest(Request):

    def __init__(self, device_id: str, index: int) -> None:
        payload = "{}:get({})".format(device_id, index)
        super().__init__(payload)

    def parse_response(self, response: str):
        resp = response.split(":").pop()
        if resp.isdecimal():
            return float(resp)
        elif resp.startswith("\"") and resp.endswith("\""):
            return resp[1:len(resp)]
        elif resp in ["true", "false"]:
            return resp == "true"
        else:
            return None

class ExecuteRequest(Request):

    def __init__(self, device_id: str, intdex: int, *args: Any) -> None:
        params_str = ''.join([(',' + str(i)) for i in args]) if len(args) > 0 else ", 0"
        payload = "{}:execute({}{})".format(device_id, intdex, params_str)
        super().__init__(payload)

    def parse_response(self, response: str):
        resp = response.split(":").pop()
        if resp.isdecimal():
            return float(resp)
        elif resp.startswith("\"") and resp.endswith("\""):
            return resp[1:len(resp)]
        elif resp in ["true", "false"]:
            return resp == "true"
        else:
            return None




import socket
from .cipher import GrentonCypher
from .requests import Request, CheckAlive

class GrentonApi:

    def __init__(self, ip: str, port: int, key: str, iv: str, timeout: int = 1) -> None:
        self._addr = (ip, port)
        self._timeout = timeout
        self._local_ip = socket.gethostbyname(socket.gethostname())
        self._cypher = GrentonCypher(key, iv)

    def send_request(self, req: Request):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.settimeout(self._timeout)

        payload = self._cypher.encrypt(bytes(req.create_request(self._local_ip), "utf-8"))

        sock.sendto(payload, self._addr)

        resp = sock.recvfrom(1024)[0]
        resp = self._cypher.decrypt(resp).decode("utf-8")
        return req.parse_response(resp)

    def checkAlive(self) -> bool:
        try:
            self.send_request(CheckAlive())
            return True
        except:
            return False

    def ip(self) -> str:
        return self._addr[0]

    def port(self) -> int:
        return self._addr[1]



from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.backends import default_backend
from base64 import b64decode

class GrentonCypher:

    def __init__(self, key: str, iv: str) -> None:
        self._key = b64decode(key)
        self._iv = b64decode(iv)
        self._padding = padding.PKCS7(128)
        self._cypher = Cipher(
            algorithms.AES(self._key), modes.CBC(self._iv), backend=default_backend()
        )

    def encrypt(self, data: bytes) -> bytes:
        encryptor = self._cypher.encryptor()
        padder = self._padding.padder()
        data = padder.update(data) + padder.finalize()
        return encryptor.update(data) + encryptor.finalize()

    def decrypt(self, data: bytes) -> bytes:
        decryptor = self._cypher.decryptor()
        unpadder = self._padding.unpadder()
        data = decryptor.update(data) + decryptor.finalize()
        return unpadder.update(data) + unpadder.finalize()
